

callback_t = Callable[[T, numpy.ndarray, Dict[str, int]], ...]


class Device:
    pattern: Pattern
    ports: numpy.ndarray
    port_names: Dict[str, int]
    callback: Optional[callback_t] = None

    def __init__(self,
                 pattern: Optional[Pattern] = None,
                 ports: Optional[numpy.ndarray] = None,
                 names: Optional[Union[Dict[str, int], List[Optional[str]]] = None,
                 callback: Optional[callback_t]= None
                 ) -> None:
        if pattern is None:
            self.pattern = Pattern()

        if ports is None:
            self.ports = {0: [0, 0,  0, 0],
                          1: [0, 0, pi, 0]}
        else:
            self.ports = numpy.array(ports)

        if callback:
            self.callback = callback
            self.callback(self, self.ports.keys())

    def __getitem__(self, key) -> numpy.ndarray:
        if isinstance(key, str):
            inds = [self.port_names[key]]
        elif hasattr(key, '__iter__'):
            inds = [self.port_names.get(k, k) for k in key]
        else:
            inds = [self.port_names.get(key, key)]
        return self.ports[inds]

    def build(self: T,
              name: str,
              other: Device,
              map_in: Dict[port_t, port_t],
              map_out: Dict[port_t, port_t],
              mirror: bool = False,
              ) -> T:
        translation, rotation = self.find_transform(other, map_in, map_out, mirror)

        pat = Pattern(name)
        pat.addsp(self.pattern)
        new = Device(pat, ports=self.ports, port_names=self.port_names, callback=self.callback)
        return new


    def plug(self, other, map_in, map_out, mirror):
        translation, rotation, pivot = self.find_transform(other, map_in, map_out, mirror)

        sp = SubPattern(other.pattern, mirrored=mirror)
        sp.rotate_around(pivot, rotation)
        sp.translate(translation)
        self.pat.subpatterns.append(sp)

        # get rid of plugged ports

        # insert remaining device ports into router port list

        with numpy.errstate(invalid='ignore'):
            self.ports[:, 2] %= 2 * pi

        if self.port_callback:
            self.port_callback(...)

    def find_transform(self, other, map_in, map_out, mirror):
        s_ports = self[map_in.keys()]
        o_ports= other[map_in.values()]

        if mirror[0]:
            o_ports[:, 1] *= -1
            o_ports[:, 2] += pi
        if mirror[1]:
            o_ports[:, 0] *= -1
            o_ports[:, 2] += pi

        s_offsets = s_ports[:, :2]
        s_angles = s_ports[:, 2]
        s_types = s_ports[:, 3]
        o_offsets = o_ports[:, :2]
        o_angles = o_ports[:, 2]
        o_types = o_ports[:, 3]

        if (r_types != d_types) & (r_types != 0) & (d_types != 0):
            #TODO warn about types

        rotations = numpy.mod(s_angles - o_angles - pi, 2 * pi)

        if not numpy.allclose(rotations[:1], rotations):
            # TODO warn about rotation

        rot_ports = rotate_ports_around(o_ports, o_offsets[0], rotations[0])       #TODO also rotate unplugged device ports

        translations = r_offsets - d_offsets
        translation = translations[0]

        if not numpy.allclose(translations[:1], translations):

        return translations[0], rotations[0]


    def as_pattern(self, name) -> Pattern:
        return self.pat.copy().rename(name)

    def as_device(self, name):
        return Device(self.as_pattern, REMAINING_NON-NAN_PORTS) #TODO


def rotate_ports_around(ports: numpy.ndarray, pivot: numpy.ndarray, angle: float) -> numpy.ndarray:
    ports[:, :2] -= pivot
    ports[:, :2] = (rotation_matrix_2d(angle) @ ports[:, :2].T).T
    ports[:, :2] += pivot
    ports[:, 2] += angle
    return ports
